@@ -1,6 +1,6 @@
 package SVN::Mirror::Ra;
 @ISA = ('SVN::Mirror');
-$VERSION = '0.55';
+$VERSION = '0.61';
 use strict;
 use SVN::Core;
 use SVN::Repos;
@@ -155,8 +155,10 @@ sub load_state {
 	if $self->{root}->node_prop ('/', join (':', 'svm:mirror', $self->{source_uuid},
 						$self->{source_path} || '/'));
 
-    die "no headrev"
-	unless defined $self->load_fromrev;
+    unless ($self->{ignore_lock}) {
+	die "no headrev"
+	    unless defined $self->load_fromrev;
+    }
     return;
 }
 
@@ -324,6 +326,7 @@ sub committed {
 	if $self->{rev_incomplete};
     $self->{headrev} = $rev;
 
+    $self->unlock ('mirror');
     print "Committed revision $rev from revision $sourcerev.\n";
 }
 
@@ -409,6 +412,7 @@ The structure of mod_lists:
     $editor->{mod_lists} = {};
     foreach ( keys %$paths ) {
         my $item = $paths->{$_};
+	s/\n/ /g; # XXX: strange edge case
         my $href;
 
         my $svn_lpath = my $local_path = $_;
@@ -439,26 +443,33 @@ The structure of mod_lists:
         if ( defined $lrev && $lrev != -1 ) {
 	    $src_lpath = $rpath;
 	    # copy within mirror anchor
-            if ($src_lpath =~ s|^\Q$self->{rsource_path}\E|$self->{target_path}|) {
+            if ($src_lpath =~ s|^\Q$self->{rsource_path}\E/|$self->{target_path}/|) {
 		# $source_node_kind is used for deciding if we need reporter later
 		my $rev_root = $self->{fs}->revision_root ($lrev);
 		$source_node_kind = $rev_root->check_path ($src_lpath);
 	    }
 	    else {
-                # The source is not in local depot.  Invalidate this
-                # copy.
-		($src_lpath, $href->{local_rev}) =
-		    $self->{cb_copy_notify}
-		    ? $self->{cb_copy_notify}->($self, $local_path, $rpath, $rrev)
-		    : (undef, undef)
-            }
+		($src_lpath, $href->{local_rev}) = (undef, undef);
+	    }
+	}
+	elsif ($rrev != -1) {
+	    # The source is not in local depot.  Invalidate this
+	    # copy.
+	    ($src_lpath, $href->{local_rev}) =
+		$self->{cb_copy_notify}
+		? $self->{cb_copy_notify}->($self, $local_path, $rpath, $rrev)
+		: (undef, undef)
         }
         @$href{qw/local_source_path source_node_kind/} =
             ( $src_lpath, $source_node_kind );
 
-        if ( /^\Q$self->{rsource_path}\E/ ) {
+	# XXX: do we still need the reanchor case?
+	# Also, the loop should not reached here if changed path is
+	# not interesting to us, skip them at the beginning the the loop
+        if ( $_ eq $self->{rsource_path} or
+	     index ("$_/", "$self->{rsource_path}/") == 0 ) {
             $editor->{mod_lists}{$svn_lpath} = $href;
-        } elsif ($href->{action} eq 'A' &&
+        } elsif (defined $lrev && $lrev != -1 && $href->{action} eq 'A' &&
 		 index ($self->{rsource_path}, "$_/") == 0) {
 	    # special case for the parent of the anchor is copied.
 	    my $reanchor = $self->{rsource_path};
@@ -487,7 +498,8 @@ The structure of mod_lists:
             my $start = $fromrev || ($self->{skip_to} ? $fromrev : $rev-1);
             my $reporter =
                 $ra->do_update ($rev, $editor->{target} || '', 1, $editor);
-            $reporter->set_path ('', $start, 0);
+	    my @lock = $SVN::Core::VERSION ge '1.2.0' ? (undef) : ();
+            $reporter->set_path ('', $start, @lock, 0);
 
             $reporter->finish_report ();
         } else {
@@ -536,8 +548,10 @@ sub get_merge_back_editor {
 	($self->_new_ra ( url => "$self->{source}$path"), "$self->{source}$path" );
 
     $self->{commit_ra} = $self->{cached_ra};
+    $self->load_fromrev;
+    my @lock = $SVN::Core::VERSION ge '1.2.0' ? (undef, 0) : ();
     return ($self->{fromrev}, SVN::Delta::Editor->new
-	    ($self->{cached_ra}->get_commit_editor ($msg, $committed)));
+	    ($self->{cached_ra}->get_commit_editor ($msg, $committed, @lock)));
 }
 
 sub switch {
@@ -570,11 +584,18 @@ sub get_latest_rev {
 	else {
 	    $headrev = $ra->get_latest_revnum
 		unless defined $headrev;
-	    $ra->get_log ([''], -1, $headrev-$offset,
+
+	    $headrev -= $offset;
+	    $ra->get_log ([''], -1, $headrev,
 			  ($SVN::Core::VERSION ge '1.2.0') ? (0) : (),
 			  0, 1,
 			  sub { $rev = $_[1] unless defined $rev});
-	    $offset*=2;
+	    if ( $offset < $headrev ) {
+		$offset*=2;
+	    }
+	    else {
+		$offset = 2;
+	    }
 	}
     }
 
@@ -587,6 +608,11 @@ sub run {
     my $self = shift;
     my $ra = $self->_new_ra;
     my $latestrev = $self->get_latest_rev ($ra);
+
+    $self->lock ('sync');
+    $self->load_fromrev;
+    $self->{headrev} = $self->{fromrev} ?
+	$self->find_local_rev ($self->{fromrev}) : $self->{fs}->youngest_rev;
     if ($self->{skip_to} && $self->{skip_to} =~ m/^HEAD(?:-(\d+))?/) {
 	$self->{skip_to} = $latestrev - ($1 || 0);
     }
@@ -597,11 +623,11 @@ sub run {
 
     print "Syncing $self->{source}".($self->_relayed ? " via $self->{rsource}\n" : "\n");
 
-    return unless $endrev == -1 || $startrev <= $endrev;
+    $self->unlock ('sync'), return
+	unless $endrev == -1 || $startrev <= $endrev;
 
     print "Retrieving log information from $startrev to $endrev\n";
 
-    $self->{headrev} = $self->{fs}->youngest_rev;
     eval {
     $ra->get_log ([''], $startrev, $endrev,
 		  ($SVN::Core::VERSION ge '1.2.0') ? (0) : (),
@@ -628,6 +654,8 @@ sub run {
     delete $self->{cached_ra};
     delete $self->{cached_ra_url};
 
+    $self->unlock ('sync');
+
     return unless $@;
     if ($@ =~ /no item/) {
 	print "Mirror source already removed.\n";
@@ -952,7 +980,6 @@ sub add_directory {
              && !$action) {
         if ($self->_contains_mod_in_path ($path)) {
             $is_copy = 1;
-            splice @_, 0, 2, $self->{mirror}{headrev};
             $method = 'open_directory';
             $self->_enter_new_copied_path ();
         } else {
@@ -981,7 +1008,6 @@ sub add_directory {
             }
         } elsif ($action eq 'M') {
             # 1d.
-            splice @_, 0, 2, $self->{mirror}{headrev};
             $method = 'open_directory';
 
             $self->_enter_new_copied_path ()
@@ -1006,11 +1032,7 @@ sub add_directory {
         die "Oh no, no more exceptions!  add_directory() failed.";
     }
 
-    if ($path eq $self->{target}) {
-        splice @_, 0, 2, $self->{mirror}{headrev}
-            if $method ne "open_directory";
-        $method = "open_directory";
-    }
+    $method = "open_directory" if $path eq $self->{target};
 
     push @{$self->{visited_paths}}, $path;
     my $tran_path = $self->_translate_rel_path ($path);
@@ -1018,12 +1040,15 @@ sub add_directory {
     $method = 'open_directory'
         if $tran_path eq $self->{mirror}{target_path};
 
+    splice @_, 0, 2, $self->{mirror}{headrev}
+	if $method eq "open_directory";
+
     my $dir_baton;
     $method = "SUPER::$method";
     $dir_baton = $self->$method($tran_path, $pb, @_);
 
     # Always 'touch' the directory, even for empty modifications.
-    $self->change_dir_prop ( $dir_baton, 'svm' => undef );
+    $self->change_dir_prop ( $dir_baton, 'svm' => undef, $pool );
 
     $self->_remove_entries_in_path ($path, $dir_baton, $pool) if $is_copy;
 
@@ -1049,7 +1074,7 @@ sub close_directory {
     $self->_leave_new_copied_path ()
         if $self->_under_latest_copypath ($path);
         
-    $self->SUPER::close_directory ($baton);
+    $self->SUPER::close_directory ($baton, @_);
 }
 
 sub close_file {
@@ -1138,15 +1163,6 @@ sub delete_entry {
     print "MirrorEditor::delete_entry($path, $rev)\n" if $debug;
     return unless $pb;
 
-    my $source_path = $self->{mirror}{rsource_path};
-    my $target_path = $self->{mirror}{target_path};
-    $source_path =~ s{^/}{};
-    if ($source_path && $path =~ m/^\Q$source_path\E/) {
-        $path =~ s/\Q$source_path\E/$target_path/;
-    } else {
-        $path = "$target_path/$path";
-    }
-
     $self->SUPER::delete_entry ($path, $self->{mirror}{headrev},
 				$pb, $pool);
 }
@@ -1165,8 +1181,8 @@ sub close_edit {
     local $SIG{INT} = 'IGNORE';
     local $SIG{TERM} = 'IGNORE';
 
-    # XXX - On Win32+fsfs, ->close_edit works but raises exceptions
-    local $@; eval { $self->SUPER::close_edit ($pool) };
+    $self->{mirror}->lock ('mirror');
+    $self->SUPER::close_edit ($pool);
 }
 
 1;
